"""Find the maximal complex in a list of cell complexes.
This is a pass-through from FIAT so that FInAT clients
(e.g. tsfc) don't have to directly import FIAT."""

from FIAT.reference_element import max_complex  # noqa: F401
